import asyncio
from django.conf import settings
from django.urls import reverse
from .default_settings import ALLOW_ACCESS_CALLBACK, REDIS_KEY_NAMESPACE


async def websocket_handler(scope, receive, send, inner_app):
    ws_path = reverse("django_live_dashboard_index")
    if scope.get("path") != ws_path:
        return await inner_app(scope, receive, send)
    init_message = await receive()
    assert init_message["type"] == "websocket.connect"
    allow_access_cb = settings.DJANGO_LIVE_DASHBOARD.get(
        "ALLOW_ACCESS_CALLBACk", ALLOW_ACCESS_CALLBACK
    )
    if allow_access_cb:
        if await allow_access_cb(scope) is not True:
            return await send({"type": "websocket.close"})
    await send({"type": "websocket.accept"})

    ws_disconnected = False

    async def task_listen_redis_events():
        from django.apps import apps

        app_config = apps.get_app_config("django_live_dashboard")
        redis = app_config.redis
        pubsub = redis.pubsub(ignore_subscribe_messages=True)
        await pubsub.subscribe(
            settings.DJANGO_LIVE_DASHBOARD.get(
                "REDIS_KEY_NAMESPACE", REDIS_KEY_NAMESPACE
            )
            + ":pubsub",
        )
        while True:
            # aredis swallows all exceptions here :( and returns None.
            # So check if disconnected after getting None
            message = await pubsub.get_message()
            if message is None:
                continue
            if not ws_disconnected:
                await send({"type": "websocket.send", "text": message["data"].decode()})
            else:
                break

    async def task_handle_received_message():
        while True:
            message = await receive()
            if message["type"] == "websocket.disconnect":
                break

    tasks = []
    redis_task = asyncio.create_task(task_listen_redis_events())
    tasks.append(redis_task)
    tasks.append(asyncio.create_task(task_handle_received_message()))
    await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
    ws_disconnected = True
    for t in tasks:
        t.cancel()



import os
import socket
from typing import Dict, List, Literal

from pydantic import BaseModel, Field


class NodeInfo(BaseModel):
    pid: str
    hostname: str


current_node = NodeInfo(pid=str(os.getpid()), hostname=socket.gethostname())


class Event(BaseModel):
    node_info: NodeInfo = Field(current_node)


class SQLQuery(Event):
    type: Literal["SQLQuery"] = Field("SQLQuery", const=True)
    sql: str
    duration: float
    db: str


class Request(Event):
    type: Literal["Request"] = Field("Request", const=True)
    path: str
    status_code: int
    time: float
    duration: float
    queries: List[SQLQuery] = []
    request_headers: Dict[str, str]
    response_headers: Dict[str, str]



import queue
import threading

from django.conf import settings
from pydantic import BaseModel
from redis import Redis

from .default_settings import REDIS_KEY_NAMESPACE, REDIS_URI
from .serializer import serialize

events_queue = queue.Queue()


def event_worker():
    redis = Redis.from_url(settings.DJANGO_LIVE_DASHBOARD.get("REDIS_URI", REDIS_URI))
    while True:
        item = events_queue.get()
        redis.publish(
            settings.DJANGO_LIVE_DASHBOARD.get(
                "REDIS_KEY_NAMESPACE", REDIS_KEY_NAMESPACE
            )
            + ":pubsub",
            item,
        )


def push_evet(event):
    if isinstance(event, BaseModel):
        events_queue.put(serialize(event.dict()))
        return
    events_queue.put(serialize(event))


event_worker_thread = threading.Thread(target=event_worker, daemon=True)
event_worker_thread.start()


async def system_info_worker():
    pass



from aredis import StrictRedis
from django.conf import settings
from .default_settings import REDIS_URI, REDIS_KEY_NAMESPACE


async def lifespan_handler(scope, receive, send, inner_app):
    from django.apps import apps

    app_config = apps.get_app_config("django_live_dashboard")

    async def wrapped_send(message):
        if message["type"] == "lifespan.startup.complete":
            app_config.asgi_started = True
        if message["type"] == "lifespan.shutdown.complete":
            app_config.asgi_shutdown = True
        await send(message)

    redis = StrictRedis.from_url(settings.DJANGO_LIVE_DASHBOARD.get("REDIS_URI", REDIS_URI))
    while True:
        message = await receive()
        if message["type"] == "lifespan.startup":
            app_config.redis = redis
            redis.hset(settings.DJANGO_LIVE_DASHBOARD.get("REDIS_KEY_NAMESPACE", REDIS_KEY_NAMESPACE) + ":nodes", )
            try:
                await inner_app(scope, receive, wrapped_send)
            except ValueError:
                pass
            if not app_config.asgi_started:
                await send({"type": "lifespan.startup.complete"})
        elif message["type"] == "lifespan.shutdown":
            try:
                await inner_app(scope, receive, wrapped_send)
            except ValueError:
                pass
            if not app_config.asgi_shutdown:
                await send({"type": "lifespan.shutdown.complete"})



import time

from django.http import HttpResponse, HttpRequest
from django.db import connections

from .worker import push_evet
from .models import Request, SQLQuery


def django_live_dashboard_asgi_middleware(app):
    from django_live_dashboard.lifespan import lifespan_handler
    from django_live_dashboard.websocket import websocket_handler

    async def wrap(scope, receive, send):
        if scope["type"] == "lifespan":
            return await lifespan_handler(scope, receive, send, app)
        if scope["type"] == "websocket":
            return await websocket_handler(scope, receive, send, app)
        else:
            await app(scope, receive, send)

    return wrap


def get_queries_data(_request):
    queries = []
    for conn in connections.all():
        for query in conn.queries:
            item_data = {
                "duration": float(query.get("time", 0)),
                "sql": query.get("sql", ""),
                "db": conn.alias,
            }
            queries.append(SQLQuery(**item_data))
    return queries


def get_request_event(now, duration, request: HttpRequest, response: HttpResponse):
    data = {
        "time": now,
        "duration": duration,
        "status_code": response.status_code,
        "path": request.build_absolute_uri(),
        "queries": get_queries_data(request),
        "request_headers": {k.lower(): v for k, v in request.headers.items()},
        "response_headers": {k.lower(): v for k, v in response.items()},
    }
    return Request(**data)


def sync_middleware(get_response, request):
    now = time.time()
    response = get_response(request)
    duration = time.time() - now
    request.duration = duration
    event = get_request_event(now, duration, request, response)
    push_evet(event)
    return response


def django_live_dashboard_middleware(get_response):
    def middleware(request):
        return sync_middleware(get_response, request)

    return middleware



from .middleware import (
    django_live_dashboard_asgi_middleware,
    django_live_dashboard_middleware,
)

__all__ = ["django_live_dashboard_asgi_middleware", "django_live_dashboard_middleware"]

default_app_config = "django_live_dashboard.apps.DjangoLiveDashboardConfig"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_live_dashboard']

package_data = \
{'': ['*'],
 'django_live_dashboard': ['static/*', 'templates/django_live_dashboard/*']}

install_requires = \
['aredis>=1.1.8,<2.0.0', 'pydantic>=1.6.1,<2.0.0', 'redis>=3.5.3,<4.0.0']

setup_kwargs = {
    'name': 'django-live-dashboard',
    'version': '0.0.1',
    'description': 'Live runtime metrics dashboard for Django',
    'long_description': None,
    'author': 'shen',
    'author_email': 'dustet@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>3.6,<4',
}


setup(**setup_kwargs)



from django.apps import AppConfig


class DjangoLiveDashboardConfig(AppConfig):
    name = "django_live_dashboard"
    verbose_name = "Django Live Dashboard"
    asgi_started = False
    asgi_shutdown = False
    redis = None
